/*
* RTSchedForObjectsThddGUI.java
*
* Created on May 24, 2005, 1:07 AM
*
*/

package scheduler;
import java.util.HashMap;
import java.util.concurrent.*;
import message.*;
import main.*;
import peer.RTPeer;

/**
*
* @author Sean Halle
 *
 *  The algorithm implemented by this scheduler is specified by the
 *   messages which it accepts and generates:
 *
 *  The run-times are organised into a hierarchy.  In this implementation
 *   (may 29th, 2005), the only hierarchy is the console, which acts as
 *   an "above" peer.
 *
 *  Large data-chunks are arranged to be transferred by the above-peer.
 *   First the above-peer picks one of the peers in a group to perform
 *   the final un-divide.  The above-peer sends "I think this data 
 *   should be un-divided by <peer>" to all peers in the group.
 *   (This is the "big chunk" message which will be referred to below.)
 *  message: [request-chain: data: take: from-above, the un-divider's ID,  
 *            info-about-data, data-owner, msgID-of-orig-msg-from-owner, 
 *            msgID] 
 *  from:    above-peer 
 *  to:      all-in-group
 *
 *  In later versions of the scheduler, negotiation may ensue to be 
 *   sure everyone agrees.  In this version, all peers are assumed to
 *   accept.
 *  The scheduler of every peer in the peer-group then performs its 
 *   algorithm to divide up the data among the peers in the group.  
 *   (according to the info each has about the other peers in its 
 *    group..  will address issues of network lag in a later version 
 *    of the scheduler).
 *  Each peer independently arrives at the same conclusion about the 
 *   division as all the others do.
 *
 *  If this peer-group is a leaf-group (which it always is in this 
 *   version of the scheduler), then each peer sends a message to each
 *   "owner" of the data saying "I get this piece" (the owner of each
 *   element of the input-set is included in the message from the 
 *   above-peer).  In this version, there is only one owner.
 *  message: [request-chain: data: give: from-below, info-about-data, 
 *            designation-msgID, msgID-of-orig, respond-to-addr-info, 
 *            msgID]
 *  from:    one of leaf-peer-group which ended up with the "big chunk"
 *  to:      owner-of-data
 *
 *  The owner replies with a "here's your piece" message that holds 
 *   the actual data, has level set to "from above", and has the 
 *   messageID it's in-response to.  (may be useful to have the 
 *   randomly-generated messageID plus originator when checking for
 *   possible attackers -- statistics of their random numbers, 
 *   sharing info about which leaf-peers don't pick up their data).
 *  message: [response: data: take: from-above, msgID-response-to, msgID]
 *  from:    owner-of-data
 *  to:      respond-to-addr taken from request-message
 *
 *  In later versions, if the owner has a piece left-over unclaimed, it
 *   initiates a "something is wrong" proceeding.
 *
 *  As each (leaf) peer receives its piece of the input-set, it hands 
 *   the data to its worker, specifying that the input-set is for the
 *   "first" dashed-unit.  
 *  message: [request: work: from-self, data-reference, dashed-unit-ID,
 *            msgID]
 *  from:    scheduler
 *  to:      worker
 *
 *  In this version of the scheduler, there is only one dashed box.
 *  If there were more, then peer-to-peer data exchanges would take
 *   place between intermediate dashed-units.
 *  When the worker is done, it hands a completion message to the 
 *   scheduler (with level set to "self")
 *  message: [response: work: from-self, msgID-in-response-to]
 *  from:    worker
 *  to:      scheduler
 *
 *  In response, the scheduler sends two messages.  
 *  It sends a completion message out to all the peers in the group.
 *   The completion message states which dashed-unit, so the others
 *   can track progress, as well as the ID of the original "big
 *   chunk" message that this completion is working towards.  This
 *   info allows the other peers to calculate where the data should
 *   go next, so all peers can keep the same image of the progress of
 *   the calculation.
 *  message: [info: completion: from-same, orig-designation-msgID, 
 *            dashed-unit-ID, msgID]
 *  from:    a-peer-in-group
 *  to:      all-peers-in-group
 *  
 *  It then sends a "here's my finished piece" to the un-divider peer
 *   specified in the original "big chunk" data message from above.
 *  message: [response: data: take: from-same, orig-designation-msgID, 
 *            msgID]
 *  from:    a-peer-in-group
 *  to:      the designated un-divider
 * 
 *  The designated un-divider applies the un-divide-dashed-unit to the
 *   pieces which come in.
 *  message: [request: work: from-self, data-reference, dashed-unit-ID,
 *            msgID]
 *  from:    scheduler
 *  to:      worker
 *
 *  Worker responds when it's done.
 *  message: [response: work: from-self, msgID-in-response-to]
 *  from:    worker
 *  to:      scheduler
 *
 *  Then scheduler sends a "here's my finished piece" to the above-peer
 *   which originally sent the "big chunk" data message.  (this 
 *   completion message has level set to "from below", and messageID
 *   it's in response to).  The data doesn't actually move yet..  only
 *   the info about it is sent.  The data will eventually move in 
 *   response to some leaf-peer somewhere requesting it.  This will 
 *   be added in future versions of the scheduler.
 *  message: [info: completion: from-below, orig-designation-msgID, 
 *            msgID]
 *  from:    designated un-divider
 *  to:      above-peer who sent original "big chunk" message
 *  
 */
public class RTSchedForObjectsThddGUI extends RTScheduler
 { 
   protected RTMsgWindow   msgWindow;
      
   /** Creates a new instance of RTSchedForObjectsThdd */
   public RTSchedForObjectsThddGUI( BlockingQueue _toClientQ, 
    BlockingQueue _toWorkerQ, BlockingQueue _toSelfQ, 
    RTPeer _parentRT ) 
    {
      super( _toClientQ, _toWorkerQ, _toSelfQ, _parentRT );

      msgWindow = new RTMsgWindow("Sched of Peer at Port: " + 
       containingPeer.getPort() );
    }
 
   /**
    * Infinite loop..  wait for something to show up from the server, 
    *  decide what action to take, send out a message through the
    *  client..
    */
   public void run()
    { RTMsgBase msgOut   = new RTMsgBase();  //bec. compiler complains
      RTMsgBase msgIn    = new RTMsgBase();  //bec. compiler complains
      boolean keepTrying = true;
      
      while( true )
       {    
            //get next thing in the queue (from server or worker)        
         msgIn = getNextMsgIn();   
         
         msgWindow.displayMsg( "Got ", msgIn );

            // figure out what kind of message it is and act accdingly
         dispatchMsg( msgIn );
                  
       } 
    } // public void run()
   

   /*
    *  Large data-chunks are arranged to be transferred by the above-
    *   peer.  First the above-peer picks one of the peers in a group 
    *   to perform the final un-divide.  The above-peer sends "I think
    *   this data should be un-divided by <peer>" to all peers in the 
    *   group.
    *   (This is the "big chunk" message referred to below.)
    *  The peers who receive this message check whether they are the 
    *   one named as the un-divider.
    *  When yes, they are the un-divider, they perform the division
    *   algorithm, using their status table to decide who gets what.
    *  When no, they are not the un-divider, they still perform the
    *   division algorithm.
    *  Either way, they all update their status tables.  They note
    *   that the peer named as un-divider has an un-division pending.
    *  Then, they all determine the meta-info for the piece of data 
    *   that they individually have decided is the piece they should 
    *   get.  They put that meta-info into a 
    *   reqChainDataGiveMsgFromAny and send it to the owner of the
    *   data.
    *  (at the moment, ignoring issues arising from network delay
    *   causing the status tables to be out-of-sync..  assuming 
    *   using a "guaranteed delivery" message-platform which can 
    *   delay things until all the messages catch up before doing
    *   the division.. in the meantime maybe do something at the 
    *   data-owner(s) to ensure consistency, each includes expected
    *   pieces others will ask for)
    *  message: [request-chain: data: take: from-above, the 
    *            un-divider's ID, info-about-data, data-owner, 
    *            msgID-of-orig-msg-from-owner, msgID] 
    *  from:    above-peer 
    *  to:      all-in-group
    */
   protected void handleReqChainDataTakeMsgFromAbove( RTMsgBase msgIn )
    {       
      msgWindow.displayMsg( "Handling ", msgIn ); 
      
         // Perform division algorithm, using the status table
         //  to base decision of who gets what on.
      int[] pieces = divideData( ((RTMsgSub)msgIn).getDataSize() );
      updateStatusTableWithNewPieces( pieces );
      RTMsgBase msgOut = makeReqChainDataGiveMsgFromAny( msgIn );
      ((RTMsgSub)msgOut).setDataSize( pieces[ containingPeer.thisRTsIndex ] );
      
      msgWindow.displayMsg( "To Client ", msgOut );
      sendMsg( msgOut );
    }

   protected int[] divideData( int size )
    {
         // no actual scheduling algorithm..  this whole thing is 
         //  just to work out the "distributed system issues"
         //  such as minimizing message traffic, synchronizing
         //  state (dealing with variable network delays), 
         //  scalability
      
      int numRTs = containingPeer.getNumRTs();
      int[] pieces = new int[ numRTs ];
      for(int i=0; i < numRTs; i++ )
       {
         pieces[i] = size/numRTs;
       }
      return pieces;
    }
  
   protected void updateStatusTableWithNewPieces( int[] pieces )
    {
      // no actual sched alg in this version.. Jun2 2005
    }
   
    /** This message arrives in response to the last link of a 
        chain-request.  It holds the actual payload.  If have
        time, will implement the data such that a single "big
        chunk" can be an aggregate of many pieces, each on a
        peer any place in the network.  
       The real message would have the data payload, but this don't.
        So, just send the size to the worker then return to wait 
        for the next message in.
     */
   protected void handleRespDataTakeMsgFromAny( RTMsgBase msgIn )
    {
      RTMsgSub msgOut = makeReqWorkMsgFromSelf( msgIn );
      msgOut.setDashedUnit( decideDashedUnitToApply( msgIn ) );
      sendMsgToWorker( msgOut );
    }

    /** This message arrives as a chained response to the original 
     *  big chunk message.  It only arrives at the undivider for 
     *  that chunk.  It's arrival means that one of the peers in the
     *  group has finished a piece.  It holds the actual data
     *  computed by that peer.  If have
     *  time, will implement the data such that a single "big
     *  chunk" can be an aggregate of many pieces, each on a
     *  peer any place in the network.
     * The data needs to have the undivider dashed-unit applied to it.
     * The real message would have the data payload, but this don't.
     *  So, just send the size to the worker then return to wait 
     *  for the next message in.
     */
   protected void handleRespDataTakeMsgFromSame( RTMsgBase msgIn )
    {
      RTMsgSub msgOut = makeReqWorkMsgFromSelf( msgIn );
      msgOut.setDashedUnit( decideDashedUnitToApply( msgIn ) );
      sendMsgToWorker( msgOut );
    }

   
   /** Expecting will only call this method when get a 
    *   resp Data Take Msg..  either from any or from same
    *  Either way, the data in could be part of an input set
    *   for any of the dashed units.
    *  For now, data from any is always for the "main" 
    *   dashed unit, and data from same is always for the
    *   "undivider" dashed unit.
    *  The "divider" dashed unit should be applied to the meta-data
    *   which arrives in the Req-Chain Data Take from Above msg.
    *   But, just doing a canned division alg, so divider-unit
    *   never executed by worker.    
    *
    */
   protected DashedUnit decideDashedUnitToApply( RTMsgBase msgIn )
    {
      // just simple alg in this version.. Jun3 2005
      switch( msgIn.getLevelFrom() )
       {
         case same: return DashedUnit.unDivider;
         case any:  return DashedUnit.main;
       }
      return null;  //should never get here, so cause error
    } 
   
    
   /** This message only arrives from the worker.  It is handing back
    *   the answer that the dashed-unit produced.
    *  No matter what, send out an info message telling all that this
    *   piece of work is done for this run-time.  They will need to
    *   know which piece of data this is completing.  They should 
    *   have inferred already that this run-time was working on this
    *   piece of data, without having been told so.
    *  Random thought..  thinking, since have a queue of work stacked
    *   up, can just wait after receiving high-penalty-rollback 
    *   messages before acting on them.  For example, rolling back 
    *   the fetch of data from the data-owner could be costly.
    *   However, starting new work can be non-costly.. if the work
    *   queue is empty, go for it, that time would be wasted (from
    *   a time-to-complete, rather than energy-to-complete).  
    *  Include in each message a time-stamp and avg-flight-time-so-far
    *   (one side receives the avg the other side sees)..  In general
    *   the flight times are not symmetric..  Anyway..  keep the avg
    *   and the variance, and maybe keep a separate avg & variance
    *   for each setting of several parameters ( which program, the
    *   network load, the data sizes, etc).  Then, simply subtract
    *   the average flight-time for the received message from the 
    *   avg flight-time of any peers whose status is a propendant of
    *   the current decision.  Wait for that difference, plus an
    *   extra amount.  The extra is set by the variance and the 
    *   penalty for being wrong, such that maximize expected outcome.
    *   IE, the probability of roll-back is set by how many std 
    *   deviations above the avg flight-time (assuming normal distr).
    *   Multiply that prob by the penalty for roll-back (which may be
    *   itself a function of how long delay), to get expected penalty
    *   as a function of delay.  Compare that to some estimate of loss
    *   due to leaving the worker empty..  when get beyond 99.99% sure
    *   any messages with earlier time-stamps have arrived, then just 
    *   queue up the work no matter what.  However, if the queue 
    *   empties before reach the desired certainty, then guaranteed
    *   no loss, so put it in.  Of several candidates, do the one 
    *   which has the earliest time-stamp.  The order of out-going
    *   completion messages needs to be the same as order in which
    *   others calculate the work should be done.  They will send
    *   resolution messages if they get msgs with time-stamps showing 
    *   a different order.  
    *   
    *  This has desirable behavior of low msg overhead, and no 
    *   performance loss, as long as have sufficient parallelism.
    *  When have a serial critical path, then the time spent waiting
    *   for messages to catch up adds to the critical path.
    *  Still, are saving the round-trip times which would be inserted
    *   for any algorithm which uses guaranteed delivery, or
    *   negotiation, or can only act after ack's back or query-and-
    *   answer..
    *  In other words, waiting a bit for messages to catch up is *no
    *   worse* than the best case of other methods.  The big question
    *   is how the variance of message flight-times changes as the 
    *   application, network load, machine load, etc change.
    *
    *  For now, just pretend to update status table, and send an info
    *   message to all the others in the group (and parent?  For use
    *   in choosing an un-divider?) 
    */
   protected void handleRespWorkMsgFromSelf( RTMsgBase _msgIn )
    { RTMsgSub msgIn = (RTMsgSub) _msgIn;
      
      updateStatusTableWithSelfWorkDone( msgIn );
      
         // make an info msg for the peers in the group
      RTMsgSub msgOut = makeInfoComplMsgFromSame( msgIn );
         // send out the info msg
      sendMsgToAllInGroup( msgOut );

         // make an info msg for the above peer
      msgOut = makeInfoComplMsgFromBelow( msgIn );
         // send to the above-peer
      sendMsgToAbovePeer( msgOut );  
      

         // now figure out what to do with the result-data.
         // If this RT is also the undivider, and the work finished
         //  was the main-dashed-unit, then send the data to
         //  the worker, specifying the undivider-dashed-unit.
         // Otherwise, if it's a main-dashed-unit completion, then
         //  package it in a message and send to undivider
         // If it's an undivider completion, check whether it's the
         //  last one of the set.  If so, the others will know, from
         //  the completion message, that the undivider is done.
         //  So only package up the completed piece and return it to
         //  the above-peer.  If it's not the last one, do nothing.
      if( msgIn.getDashedUnit() != DashedUnit.unDivider )
       {
         msgWindow.displayText("main dashed unit was completed\n");
         if( containingPeer.getPort() == msgIn.getUndivider() )
          {
               // make a msg which packages the data and gives it to
               //  the worker, with the undivider as the dashed-unit.
            msgOut = makeReqWorkMsgFromSelf( msgIn );
            msgOut.setDashedUnit( DashedUnit.unDivider );
            sendMsgToWorker( msgOut );
          }
         else
          {    // pkg data into message and send to the undivider
            msgOut = makeRespDataTakeMsgFromSame( msgIn );
            sendMsg( msgOut );
          }
       }
      else // is the result of an application of the undivider
       {
         msgWindow.displayText("an undivider unit was completed\n");
            // Check whether this is the last piece of the 
            //  original "big chunk" to go through the undivider
         if( isLastChunkToUndivide( msgIn ) )
          {        

            msgWindow.displayText("last chunk, package up data");
            
            // send the data to the above peer.. msg from worker
            //  should have the correct size.
            msgOut = makeRespDataTakeMsgFromBelow( msgIn );
            sendMsgToAbovePeer( msgOut );
          }
       }
      
    }
   
   // note: overrode hashCode and equals for RTMsgID..  not sure if
   //  did it in a way that hashMap will still work right.. 
   HashMap theUndividerCounters = new HashMap();
   RTMsgID origRTMsgID = new RTMsgID();
   RTCounter theCounter;
   protected boolean isLastChunkToUndivide( RTMsgSub msgIn )
    {
      // increment the count of how many chunks have been undivided
      //  for the originial msg ID + orig msg whoFrom
      // Next, check if the count is == the number of peers in group
      //  if so, remove the counter for this origMsgID & origWhoFrom
      //  and return true, else return false
      origRTMsgID.setLocalMsgID( msgIn.getOrigLocalMsgID() );
      origRTMsgID.setWhoFrom( msgIn.getOrigWhoFrom() );
      
      
         // get counter for the big-chunk.  When none exists 
         //  make one
      theCounter = 
       (RTCounter) theUndividerCounters.get( origRTMsgID );
      if( theCounter == null )
       {
         //msgWindow.displayText("making new counter");
         theCounter = new RTCounter();
         theUndividerCounters.put( origRTMsgID, theCounter );         
       }
         // Next, incr the counter
      theCounter.increment();
      if( theCounter.hasCount( containingPeer.getNumRTs() ) )
       {    //remove the counter's mapping (should garbage collect)
         theUndividerCounters.remove( origRTMsgID );
         msgWindow.displayText("done undividing");
         return true;
       }
      else
       { 
         //msgWindow.displayText( theCounter.toString() );
         return false;
       }
    }
      
   protected void updateStatusTableWithSelfWorkDone( RTMsgBase msgIn )
    {
      // no actual sched alg in this version.. Jun2 2005
    }

   
   /**
    *  Override the standard error handler, so that it uses the
    *   message window.  The message window is created here in 
    *   the sub-class.  
    *   Any time the combination of types
    *   is invalid, end up here..  print a message to the msgWindow
    */
   protected void handleMsgError( RTMsgBase msgIn )
    {
      msgWindow.displayMsg ("An erroneous msg came in", msgIn );
    }
   
 } // public class RTSchedForObjectsThddGUI extends RTScheduler
