/*
 * RTScheduler.java
 *
 * Created on May 27, 2005, 2:02 AM
 *
 */

package scheduler;

import java.io.*;
import java.net.*;
import java.util.concurrent.*;
import java.util.Random.*;
import java.util.*;
import message.*;
import server.*;
import worker.*;
import client.*;
import main.*;
import peer.RTPeer;


/**
 *
 * @author Kevinsean Halle
 */
public class RTScheduler extends Thread
 { protected BlockingQueue toClientQ;
   protected BlockingQueue toSelfQ;
   protected BlockingQueue toWorkerQ;
   protected RTPeer        containingPeer;
   protected int           thisRTsNextMsgID = 0;
   protected Random        randNumGen;
   
   /**
    * Creates a new instance of RTScheduler. 
    *   The only allowed way to create a scheduler is by giving the
    *   constructor a reference to each queue in the run-time, and
    *   a reference to the RTPeer object that the scheduler resides
    *   within.
    */
   public RTScheduler( 
    BlockingQueue _toClientQ, BlockingQueue _toWorkerQ, 
    BlockingQueue _toSelfQ, RTPeer _containingPeer ) 
    {
      toClientQ               = _toClientQ;
      toWorkerQ               = _toWorkerQ;
      toSelfQ                 = _toSelfQ;
      containingPeer                = _containingPeer;
    
      randNumGen        = new Random( System.nanoTime() );
    }
  


   
   /*****************  The Tree of Dispatchers ***************/
   
   /** A message comes in, and goes through this decision tree
       to determine which handler to call for it.

      Override the handlers in the sub-classes to get behavior
       specific to each scheduling algorithm

      Could make a handler for every possible combination, and
       make their default behaviors be to give an error message,
       but less work to do in the dispatch tree..
      So,
      Change the dispatchers and add handlers as implement new
        message types, or override the changed portions of the
        dispatch tree and add sub-class-specific handlers..
    */   
   protected void dispatchMsg( RTMsgBase msgIn )
    { 
      switch( msgIn.getMsgType1() ) 
       {
         case Info:           dispatchInfoMsg    ( msgIn ); break;
         case Request:        dispatchReqMsg     ( msgIn ); break;
         case RequestChain:   dispatchReqChainMsg( msgIn ); break;
         case Response:       dispatchRespMsg    ( msgIn ); break;
       }
    }

   protected void dispatchInfoMsg( RTMsgBase msgIn )
    { 
      switch( msgIn.getMsgType2() ) 
       {
         case completion: dispatchInfoComplMsg( msgIn );    break;
         case data: handleMsgError( msgIn );                break;
         case work: handleMsgError( msgIn );                break;
       }
    }

   protected void dispatchReqMsg( RTMsgBase msgIn )
    { 
      switch( msgIn.getMsgType2() ) 
       {
         case completion: handleMsgError( msgIn );          break;
         case data: handleMsgError( msgIn );                break;
         case work: handleMsgError( msgIn );                break;
       }
    }

   protected void dispatchReqChainMsg( RTMsgBase msgIn )
    { 
      switch( msgIn.getMsgType2() ) 
       {
         case completion: handleMsgError( msgIn );          break;
         case data: dispatchReqChainDataMsg( msgIn );       break;
         case work: handleMsgError( msgIn );                break;
       }
    }

   protected void dispatchRespMsg( RTMsgBase msgIn )
    { 
      switch( msgIn.getMsgType2() ) 
       {
         case completion: handleMsgError( msgIn );          break;
         case data: dispatchRespDataMsg( msgIn );           break;
         case work: dispatchRespWorkMsg( msgIn );           break;
       }
    }

   
   protected void dispatchInfoComplMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getLevelFrom() )
       {
         case above: handleMsgError( msgIn );               break; 
         case below: handleInfoComplMsgFromBelow( msgIn );  break;
         case same:  handleInfoComplMsgFromSame( msgIn );   break;
         case self:  handleMsgError( msgIn );               break;
         case any:   handleMsgError( msgIn );               break;
       }
    }

   
   protected void dispatchReqChainDataMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getMsgType3() )
       {
         case give: dispatchReqChainDataGiveMsg( msgIn );   break;
         case take: dispatchReqChainDataTakeMsg( msgIn );   break;
       }
    }

   protected void dispatchRespDataMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getMsgType3() )
       {
         case give: handleMsgError( msgIn );                break;
         case take: dispatchRespDataTakeMsg( msgIn );       break;
       }
    }

   protected void dispatchRespWorkMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getLevelFrom() )
       {
         case above: handleMsgError( msgIn );               break; 
         case below: handleMsgError( msgIn );               break;
         case same:  handleMsgError( msgIn );               break;
         case self:  handleRespWorkMsgFromSelf( msgIn );    break;
         case any:   handleMsgError( msgIn );               break;
       }
    }

   
   protected void dispatchReqChainDataGiveMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getLevelFrom() )
       {
         case above: handleMsgError( msgIn );               break; 
         case below: handleMsgError( msgIn );               break;
         case same:  handleMsgError( msgIn );               break;
         case self:  handleMsgError( msgIn );               break;
         case any:   handleReqChainDataGiveMsgFromAny( msgIn ); break;
       }
    }

   
   protected void dispatchReqChainDataTakeMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getLevelFrom() )
       {
            // expect from same, and below to be added when add
            //  hierarchy
         case above: handleReqChainDataTakeMsgFromAbove( msgIn ); break; 
         case below: handleMsgError( msgIn );               break;
         case same:  handleMsgError( msgIn );               break;
         case self:  handleMsgError( msgIn );               break;
         case any:   handleMsgError( msgIn );               break;
       }
    }

   
   protected void dispatchRespDataTakeMsg( RTMsgBase msgIn )
    {
      switch( msgIn.getLevelFrom() )
       {
         case above: handleMsgError( msgIn );                 break; 
         case below: handleRespDataTakeMsgFromBelow( msgIn ); break;
         case same:  handleRespDataTakeMsgFromSame( msgIn );  break;
         case self:  handleMsgError( msgIn );                 break;
         case any:   handleRespDataTakeMsgFromAny( msgIn );   break;
       }
    }

   
/**********************  The Message Handlers  *********************/   
 
   // override these in sub-classes to define scheduling-alg
   //  specific behaviors..

   /**
    *  Large data-chunks are arranged to be transferred by the above-peer.
    *   First the above-peer picks one of the peers in a group to perform
    *   the final un-divide.  The above-peer sends "I think this data 
    *   should be un-divided by <peer>" to all peers in the group.
    *   (This is the "big chunk" message which will be referred to below.)
    *  message: [request-chain: data: take: from-above, the 
    *            un-divider's ID, info-about-data, data-owner, 
    *            msgID-of-orig-msg-from-owner, msgID] 
    *  from:    above-peer 
    *  to:      all-in-group
    */
   protected void handleReqChainDataTakeMsgFromAbove( RTMsgBase msgIn )
    {
      
    }
   
   
   /** this should only get handed a ReqChainDataTakeMsg
        from any one of above, below, or same.  The very
        first to start the chain has to insert the "orig"
        field-values explicitly into a dummy message.
    */
   protected RTMsgSub makeReqChainDataTakeMsgFromAbove( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      newMsg.setLevelFrom( Level.above );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );
      newMsg.setWhoFrom( containingPeer.getPort() );
      
      return newMsg;
    }

   protected RTMsgSub makeStarterReqChainDataTakeMsg( )
    {
         // this creates a message from scratch, to start the
         //  chain..  all the rest in the chain are cloned from
         //  preceding message.
      RTMsgSub newMsg = new RTMsgSub();
      newMsg.setMsgType1   ( MsgType1.RequestChain );
      newMsg.setMsgType2   ( MsgType2.data );
      newMsg.setMsgType3   ( MsgType3.take );
      newMsg.setDataOwner  ( containingPeer.getPort() );
      newMsg.setOrigLocalMsgID  ( thisRTsNextMsgID );   // don't incr!
      newMsg.setOrigWhoFrom( containingPeer.getPort() );
      
      return newMsg;
    }

   
   
   /** 
    *  In later versions of the scheduler, negotiation may ensue to be 
    *   sure everyone agrees.  In this version, all peers are assumed to
    *   accept.
    *  The scheduler of every peer in the peer-group then performs its 
    *   algorithm to divide up the data among the peers in the group.  
    *   (according to the info each has about the other peers in its 
    *    group..  will address issues of network lag in a later version 
    *    of the scheduler).
    *  Each peer independently arrives at the same conclusion about the 
    *   division as all the others do.
    *
    *  If this peer-group is a leaf-group (which it always is in this 
    *   version of the scheduler), then each peer sends a message to each
    *   "owner" of the data saying "I get this piece" (the owner of each
    *   element of the input-set is included in the message from the 
    *   above-peer).  In this version, there is only one owner.
    *  message: [request-chain: data: give: from-below, info-about-data, 
    *            designation-msgID, msgID-of-orig, respond-to-addr-info, 
    *            msgID]
    *  from:    one of leaf-peer-group which ended up with the "big chunk"
    *  to:      owner-of-data
    */
   protected void handleReqChainDataGiveMsgFromAny( RTMsgBase msgIn )
    {
      
    }

   /**
       this should only get handed a ReqChainDataTakeMsg
        from above.  It is the last in the chain, and 
        cannot start a chain.
    */
   protected RTMsgBase makeReqChainDataGiveMsgFromAny( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      newMsg.setMsgType3( MsgType3.give );
      
         // there's no telling the relative levels of data
         //  owner and the leaf-RT requesting the data (they're
         //  both leaves, but leaves can appear at different
         //  levels in the hierarchy (big machines vs little)
      newMsg.setLevelFrom( Level.any );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );
      newMsg.setWhoFrom( containingPeer.getPort() );
      newMsg.setPortTo( ((RTMsgSub)msgIn).getDataOwner() );
      
      return newMsg;
    }
   
   
   /**
    *  The owner replies with a "here's your piece" message that holds 
    *   the actual data, no telling what level the owner is on, so 
    *   has level set to "any", and has the 
    *   messageID it's in-response to.  (may be useful to have the 
    *   randomly-generated messageID plus originator when checking for
    *   possible attackers -- statistics of their random numbers, 
    *   sharing info about which leaf-peers don't pick up their data).
    *  message: [response: data: take: from-any, msgID-response-to, msgID]
    *  from:    owner-of-data
    *  to:      respond-to-addr taken from request-message
    */
   protected void handleRespDataTakeMsgFromAny( RTMsgBase msgIn )
    {
      
    }

   /**
       this should only get handed a ReqChainDataGiveMsg
       from any.  It is the response to the chain.
    */
   protected RTMsgSub makeRespDataTakeMsgFromAny( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      
      newMsg.setMsgType1( MsgType1.Response );
      newMsg.setMsgType2( MsgType2.data );
      newMsg.setMsgType3( MsgType3.take );
      
         // there's no telling the relative levels of data
         //  owner and the leaf-RT requesting the data (they're
         //  both leaves, but leaves can appear at different
         //  levels in the hierarchy (big machines vs little)
      newMsg.setLevelFrom( Level.any );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );
      newMsg.setWhoFrom( containingPeer.getPort() );
      newMsg.setPortTo( ((RTMsgSub)msgIn).getWhoFrom() );
      
      return newMsg;
    }
   
   
   /*
    *  In later versions, if the owner has a piece left-over unclaimed, it
    *   initiates a "something is wrong" proceeding.
    */
   
   
   
   /**
    *  As each (leaf) peer receives its piece of the input-set, it hands 
    *   the data to its worker, specifying that the input-set is for the
    *   "first" dashed-unit.
    *
    * Also:  
    *  The designated un-divider applies the un-divide-dashed-unit to the
    *   pieces which come in.
    *
    *  message: [request: work: from-self, data-reference, dashed-unit-ID,
    *            msgID]
    *  from:    scheduler
    *  to:      worker
 
       this may get handed a RespDataTakeMsg from any or from 
        same, or a Resp Work from self, when the undivider itself
        finishes a piece.  This message happens when the
        data is here and ready to be worked on.  This message
        initiates the work in the worker.  The real message
        would have a reference to the location of the data in
        memory, but this version doesn't have any real data.
    *
    */
   protected RTMsgSub makeReqWorkMsgFromSelf( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      
      newMsg.setMsgType1( MsgType1.Request );
      newMsg.setMsgType2( MsgType2.work );
      newMsg.setLevelFrom( Level.self );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );  //internal msg: incr?
      newMsg.setWhoFrom( containingPeer.getPort() ); //irrelevant
      newMsg.setPortTo( (short)0000 );
      
      return newMsg;
    }
      
   
   /** One of these messages gets received when the worker has 
    *   finished applying a dashed-unit to an input-set.
    *  In this version of the scheduler, there is only one dashed unit
    *   for application-code, plus a divider dashed-unit (unused in 
    *   this version), and an un-divider dashed-unit.
    *  If there were more, then peer-to-peer data exchanges would take
    *   place between intermediate dashed-units.
    *  When the worker is done, it hands one of these message to the 
    *   scheduler (with level set to "self")
    *  message: [response: work: from-self, msgID-in-response-to]
    *  from:    worker
    *  to:      scheduler
    */
   protected void handleRespWorkMsgFromSelf( RTMsgBase msgIn )
    {
      
    }

   
   
   /**
    *  In response, the scheduler sends two messages.  
    *  It sends a completion message out to all the peers in the group.
    *   The completion message states which dashed-unit, so the others
    *   can track progress, as well as the ID of the original "big
    *   chunk" message that this completion is working towards.  This
    *   info allows the other peers to calculate where the data should
    *   go next, so all peers can keep the same image of the progress
    *   of the calculation.  It then sends either a message to the
    *   worker, telling it to apply the undivider dashed-unit to the
    *   newly completed data, or sends the data to the designated
    *   unidivider run-time, depending on whether this run-time is 
    *   the designated un-divider.
    *
    * Also:
    *  this comes in when the undivider completes an un-divide piece.
    *   The undivider sent this out to 
    *   let others in group know so that they can update their table 
    *   with this RT's new
    *   status. (only sent from designated un-divider)
    *
    *  message: [info: completion: from-same, orig-designation-msgID, 
    *            dashed-unit-ID, msgID]
    *  from:    a-peer-in-group
    *  to:      all-peers-in-group
    */
   protected void handleInfoComplMsgFromSame( RTMsgBase msgIn )
    {
      
    }


   /** This should only get handed a Resp Work Msg
        from self.  This message happens when the
        computation is complete.  It
        causes the other peers to update their status tables.
        It needs to indicate *which* work has completed.
       The others will only have inferred what data this RT was
        working on.  So, this info message must include enough
        info to allow the other RTs to disambiguate.  If they
        detect an inconsistency they will initiate a resolution
        process (in later versions).
    */
   protected RTMsgSub makeInfoComplMsgFromSame( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      
      newMsg.setMsgType1( MsgType1.Info );
      newMsg.setMsgType2( MsgType2.completion );
      newMsg.setLevelFrom( Level.same );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );  //internal msg: incr?
      newMsg.setWhoFrom( containingPeer.getPort() ); //irrelevant
      newMsg.setPortTo( (short)0000 );
      
      return newMsg;
    }

   
   
   /** Get this anytime one of the peers in the group below finishes
    *   computing something.  It lets the above-peer track the status
    *   of all the peers.  It uses this status to decide which to name
    *   as the un-divider.
    */
   protected void handleInfoComplMsgFromBelow( RTMsgBase msgIn )
    {
 
    }

   
   /** sent to above-peer so it can also track the status of all
    *   peers in group below it.
    */
   protected RTMsgSub makeInfoComplMsgFromBelow( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = makeInfoComplMsgFromSame( msgIn );

      newMsg.setLevelFrom( Level.below );
      newMsg.setPortTo( containingPeer.getPortOfAbovePeer() );
      
      return newMsg;
    }

   
   
   /**
    *  A "here's my finished piece". It arrives to the un-divider peer
    *   specified in the original "big chunk" data message from above.
    *  message: [response: data: take: from-same, orig-designation-msgID, 
    *            msgID]
    *  from:    a-peer-in-group
    *  to:      the designated un-divider
    */ 
   protected void handleRespDataTakeMsgFromSame( RTMsgBase msgIn )
    {
      
    }

   
   /** This is generated in response to a Resp Work Msg from Self
    *   It packages the newly-completed data-result, and gets sent
    *   to the un-divider, which will assemble the pieces.
    *  Need to have the un-divider which was specified in the
    *   original chain-req data take from above message.  So, want
    *   to make sure the un-divider makes it to the worker and back.
    */
   protected RTMsgSub makeRespDataTakeMsgFromSame( RTMsgBase msgIn )
    {
      RTMsgSub newMsg = ((RTMsgSub)msgIn).clone();
      
      newMsg.setMsgType1( MsgType1.Response );
      newMsg.setMsgType2( MsgType2.data );
      newMsg.setMsgType3( MsgType3.take );
      newMsg.setLevelFrom( Level.same );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );
      newMsg.setWhoFrom( containingPeer.getPort() ); 
      newMsg.setPortTo( ((RTMsgSub)msgIn).getUndivider() );
      
      return newMsg;
    }

   
   /**
    *  Then scheduler sends a "here's my finished piece" to the above-peer
    *   which originally sent the "big chunk" data message.  (this 
    *   completion message has level set to "from below", and messageID
    *   it's in response to).  The data doesn't actually move yet..  only
    *   the info about it is sent.  The data will eventually move in 
    *   response to some leaf-peer somewhere requesting it.  This will 
    *   be added in future versions of the scheduler.
    *  message: [info: completion: from-below, orig-designation-msgID, 
    *            msgID]
    *  from:    designated un-divider
    *  to:      above-peer who sent original "big chunk" message
    */  
   protected void handleRespDataTakeMsgFromBelow( RTMsgBase msgIn )
    {
      
    }

   /** Gets handed a Compl | Work message..  which is the completion
    *   of the undivider-unit on the last piece of the "big chunk".
    *  This message gets sent to the above-peer. 
    *
    */
   protected RTMsgSub makeRespDataTakeMsgFromBelow( RTMsgBase _msgIn )
    { RTMsgSub msgIn = (RTMsgSub) _msgIn;
      RTMsgSub newMsg = msgIn.clone();
      
      newMsg.setMsgType1( MsgType1.Response );
      newMsg.setMsgType2( MsgType2.data );
      newMsg.setMsgType3( MsgType3.take );
      newMsg.setLevelFrom( Level.below );
      newMsg.setLocalMsgID ( thisRTsNextMsgID++ );
      newMsg.setWhoFrom( containingPeer.getPort() ); 
      newMsg.setPortTo( msgIn.getOrigWhoFrom() );
      
      return newMsg;
    }    

   
   /*
    *  The standard error handler.  Any time the combination of types
    *   is invalid, end up here..  print a message to the msgWindow
    */
   protected void handleMsgError( RTMsgBase msgIn )
    {
    }
   
   
   /********************  Helper Methods *****************/

         // for reuse by helper methods
   protected boolean       keepTrying = true;
   protected RTMsgBase         msg;
   
   protected void sendMsgToAllInGroup( RTMsgBase msgToSend )
    {
      RTMsgBase msgOut = msgToSend;
      
      Iterator<Short> it = 
       containingPeer.getPortsOfOtherRTsInGroup().iterator();
      
         // send the message have, before making copies..
         //  this ensures use every msgID with no skips
      Short port = it.next();
      msgOut.setPortTo( port.shortValue() );
      sendMsg( msgOut );
      
      while (it.hasNext()) 
       { port = it.next();
         msgOut = msgToSend.clone();
         msgOut.setPortTo( port.shortValue() );
         msgOut.setLocalMsgID( thisRTsNextMsgID++ );

         sendMsg( msgOut );
       }
      
    }
   
   protected RTMsgBase getNextMsgIn( )
    {    // send by placing on the queue to the client
      keepTrying = true;
      while (keepTrying == true)
       { try
          { msg = (RTMsgBase) toSelfQ.take();
            keepTrying = false;
          }
         catch (InterruptedException e)
          {    // interrupted waiting on queue.. so keep trying 
            keepTrying = true;
          }
       }
      return msg;
    }
   
   protected void sendMsg( RTMsgBase msgToSend )
    {    // send by placing on the queue to the client
      keepTrying = true;
      while (keepTrying == true)
       { try
          { toClientQ.put(msgToSend);
            keepTrying = false;
          }
         catch (InterruptedException e)
          {    // means got interrupted while waiting to put into queue.. 
               // leave keepTrying true.. and try again!
            keepTrying = true;
          } 
       }
    }
   
   protected void sendMsgToWorker( RTMsgBase msgToSend )
    {    // send by placing on the queue to the client
      keepTrying = true;
      while (keepTrying == true)
       { try
          { toWorkerQ.put(msgToSend);
            keepTrying = false;
          }
         catch (InterruptedException e)
          {    // means got interrupted while waiting to put into queue.. 
               // leave keepTrying true.. and try again!
            keepTrying = true;
          } 
       }
    }
   
   
   protected void sendMsgToAbovePeer( RTMsgBase msgToSend )
    {    // be sure dest is correct (helper)
      msgToSend.setPortTo( containingPeer.getPortOfAbovePeer() );

      sendMsg( msgToSend );
    }
   
 }
